/*
QStreamer, a camera-capture library version 1.0.0, Feb 6, 2010

Copyright (C) 2010 David C. McCallum

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from
the use ofthis software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must
      not claim that you wrote the original software. If you use this
      software in a product, an acknowledgment in the product
      documentation would be appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must
      not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source
      distribution.

David McCallum dmccallum9@gmail.com
*/

#include "QStreamer.hpp"
#include <QDir>

using namespace std;

int QStreamer::dirID = 0;

QStreamer::QStreamer(void)
{
    //### Initialize member variables ###
    deviceName = "/dev/video0";
    ramdriveDir = "/dev/shm";
    framesPerSecond = 15;
    fileType = "ppm";
    resolution = "320x240";
    errMsg = "allOK";
    myID = dirID;
    streamerProc = new QProcess(this);
    programName = "streamer";
    streamerStarted = false;
    haveTargetImage = false;

    //### Specify a sub-directory to store image files ###
    streamerDir = "/QStreamer/";
    streamerDir.append(QString::number(myID, 10));
    streamerDir.append("_images");

    //### Connect the timer ###
    connect(&logicTimer, SIGNAL(timeout()),
            this, SLOT(updateLogic()));
    logicTimer.setInterval(UPDATE_TIME);

    dirID++;
}//constructor

QStreamer::~QStreamer(void)
{
    stopStream();

    delete streamerProc;
    streamerProc = NULL;
}//destructor

bool QStreamer::startStream(QString newDeviceName,
                            int newFPS,
                            unsigned int xRes,
                            unsigned int yRes)
{
    QString filePath;//full path to the first image that should be saved by streamer
    bool result = true;//false if errors occurred

    //### Clear the previous arguments to the streamer program ###
    streamerArgs.clear();

    //### Set up the path and filenames to write to ###
    filePath.append(ramdriveDir);
    filePath.append(streamerDir);
    filePath.append("/img000000.ppm");
    streamerArgs << "-o" << filePath;

    //### Set up the file type ###
    streamerArgs << "-f" << fileType;

    //### Set up the device name ###
    if (result)
    {
        if (newDeviceName.size() > 0)
        {
            deviceName = newDeviceName;
            streamerArgs << "-c" << deviceName;
        }//if size()
        else
        {
            errMsg = "Invalid device name.";
            result = false;
        }//else size()
    }//if result

    //### Set up the resolution ###
    if (result)
    {
        if ((xRes > 0) && (yRes > 0))
        {
            resolution = QString::number(xRes, 10) + "x" + QString::number(yRes, 10);
            streamerArgs << "-s" << resolution;
        }//if xRes && yRes
        else
        {
            errMsg = "Invalid resolution.";
            result = false;
        }//else xRes && yRes
    }//if result

    //### Set a time limit on capture ###
    if (result)
        streamerArgs << "-t" << "06:00:00";

    //### Set the FPS ###
    if (result)
    {
        if ((newFPS > 0) && (newFPS <= 100))
        {
            framesPerSecond = newFPS;
            streamerArgs << "-r" << QString::number(framesPerSecond, 10);
        }//if newFPS && newFPS
        else
        {
            errMsg = "Invalid FPS.";
            result = false;
        }//else newFPS && newFPS
    }//if result

    //### Stop the video stream, if it was previously started ###
    if (result)
    {
        if (!stopStream())
            result = false;
    }//if result

    //### Create a subdirectory in the ramdrive, if necessary ###
    if (result)
    {
        if (!makeSubdirectory())
            result = false;
    }//if result

    //### Run the streamer program ###
    if (result)
    {
        cout << programName.toStdString() << " ";
        for (int i = 0; i < streamerArgs.size(); i++)
            cout << " " << streamerArgs.at(i).toStdString();
        cout << endl;

        streamerProc->start(programName, streamerArgs, QIODevice::ReadWrite);
        logicTimer.start();
        streamerStarted = true;
    }//if result

    return result;
}//start

bool QStreamer::stopStream(void)
{
    QString currentFile;//the filename of the current image to delete
    QFileInfoList imageList;//the list of image files in the ramdrive
    bool result = true;//false if we couldn't stop the streamer program

    if (streamerStarted)
    {
        //### Stop the streamer process ###
        streamerProc->terminate();
        if (!streamerProc->waitForFinished(-1))
        {
            cout << "Couldn't terminate " << programName.toStdString() << endl;
            cout << "Qt Error: " << streamerProc->errorString().toStdString() << endl;
            cout << "Trying to kill... ";

            streamerProc->kill();
            if (streamerProc->waitForFinished(-1))
                cout << " success!" << endl;
            else
            {
                errMsg = "Couldn't stop streamer! If it's still running, your ramdrive is probably filling up with images. ";
                result = false;
            }//else waitForFinished()
        }//if !waitForFinished()

        //### Delete any leftover images ###
        imageList = imageDir.entryInfoList();

        for (int i = 0; i < imageList.size(); ++i)
        {
            currentFile = imageList.at(i).fileName();

            if (!imageDir.remove(currentFile))
            {
                errMsg = "streamer stopped, but couldn't clear old images from the ramdrive";
                result = false;
            }//if !remove()
        }//for i

        //### Stop the logic updates ###
        logicTimer.stop();
        streamerStarted = false;
        targetImage.clear();
        haveTargetImage = false;
        targetImage.clear();
    }//if streamerStarted

    return result;
}//stopStream

void QStreamer::updateLogic(void)
{
    QString currentFile;//the filename of the current image to delete
    QFileInfoList imageList;//the list of image files in the ramdrive
    QString newTargetImage;//udpate the target image after the current one was loaded

    //### Load the image from the previous time updateLogic was called ###
    if (haveTargetImage)
    {
        if (loadImage(targetImage))
            haveTargetImage = false;
    }//if haveTargetImage

    //### If we loaded the target image, update it ###
    if (!haveTargetImage)
    {
        //Delete all but the most recent image
        imageList = imageDir.entryInfoList();

        for (int i = 0; i < imageList.size() - 1; ++i)
        {
            currentFile = imageList.at(i).fileName();

            if (!imageDir.remove(currentFile))
            {
                cout << "QStreamer couldn't remove " << currentFile.toStdString() << endl;
                cout << "Aborting streamer!" << endl;
                stopStream();
            }//if !remove()
        }//for i

        //Update the target image
        if (imageList.size() >= 1)
        {
            newTargetImage = imageList.at(imageList.size() - 1).fileName();
            if (targetImage != newTargetImage)
            {
                targetImage = newTargetImage;
                haveTargetImage = true;
            }//if targetImage
        }//if size()
    }//if !haveTargetImage
}//updateLogic

bool QStreamer::loadImage(const QString &imageFile)
{
    QImage latestCam;//the webcam image to load
    QString imagePath;//the path to our image directory
    bool result = true;//false on load error

    imagePath = imageDir.absolutePath() + "/" + imageFile;

    if (latestCam.load(imagePath, fileType.toAscii()))
        emit newImage(latestCam);
    else
        result = false;

    return result;
}//loadImage

bool QStreamer::makeSubdirectory(void)
{
    QString myDirectory;//the directory where I'll save my files
    bool result = true;//false if we couldn't create the directory

    myDirectory = QString::number(myID, 10);
    myDirectory.append("_images");

    //### Move to the ram drive ###
    if (!imageDir.cd(ramdriveDir))
    {
        errMsg = "Couldn't access directory: ";
        errMsg.append(ramdriveDir);
        result = false;
    }//if !cd()

    //### Create the QStreamer directory, if necessary ###
    if ((result) && (!imageDir.exists("QStreamer")))
    {
        if (!imageDir.mkdir("QStreamer"))
        {
            errMsg = "Couldn't create streamer directory: QStreamer";
            result = false;
        }//if !mkdir
    }//if result && !exists()

    //### Move to the QStreamer directory ###
    if (result)
    {
        if (!imageDir.cd("QStreamer"))
        {
            errMsg = "Couldn't access directory: QStreamer";
            result = false;
        }//if !cd()
    }//if result

    //### Create my image directory, if necessary ###
    if ((result) && (!imageDir.exists(myDirectory)))
    {
        if (!imageDir.mkdir(myDirectory))
        {
            errMsg = "Couldn't create directory: ";
            errMsg.append(myDirectory);
            result = false;
        }//if !mkdir
    }//if result && !exists()

    //### Move to my image directory and set filter ###
    if (imageDir.cd(myDirectory))
    {
        imageDir.setFilter(QDir::Files | QDir::NoSymLinks | QDir::NoDotAndDotDot);
        imageDir.setSorting(QDir::Name);
    }//if cd()
    else
    {
        errMsg = "Couldn't access directory: ";
        errMsg.append(myDirectory);
        result = false;
    }//else cd()

    return result;
}//makeSubdirectory

void QStreamer::setRamdriveDir(QString newName)
{
    ramdriveDir = newName;
}//setRamdriveDir

QString QStreamer::getErrMsg(void)
{
    return errMsg;
}//getErrMsg
