/*
QStreamer, a camera-capture library version 1.0.0, Feb 6, 2010

Copyright (C) 2010 David C. McCallum

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from
the use ofthis software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must
      not claim that you wrote the original software. If you use this
      software in a product, an acknowledgment in the product
      documentation would be appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must
      not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source
      distribution.

David McCallum dmccallum9@gmail.com
*/

#ifndef STREAMCAM_CLASS
#define STREAMCAM_CLASS

#include <QObject>
#include <QImage>
#include <QTimer>

#include <iostream>
#include <vector>
#include <QStringList>
#include <QProcess>
#include <QDir>

class QStreamer : public QObject
{
    Q_OBJECT

    public:
        //constructor
        QStreamer(void);

        //destructor
        ~QStreamer(void);

        //Start getting images from the specified device. Stops the existing stream,
        //if one exists.
        //Returns false if errors occurred, and sets errMsg appropriately.
        bool startStream(QString newDeviceName,
                         int newFPS,
                         unsigned int xRes,
                         unsigned int yRes);

        //Stop getting images.  Tries to terminate the streamer, and then kills it if necessary.
        //Returns false if we couldn't stop it.
        bool stopStream(void);

        //Change the directory of the system's ramdrive, should normally be /dev/shm
        //Takes effect the next time startStream is called.
        void setRamdriveDir(QString newName);

        //returns a description of the most recent error
        QString getErrMsg(void);

        //The number of arguments we'll provide to streamer
        static const int NUM_ARGUMENTS = 13;

        //The number of milliseconds to wait before checking the ramdrive for images
        static const int UPDATE_TIME = 10;

    signals:
        //emitted when we get a new image from the camera
        void newImage(QImage &theImage);

    private slots:
        //periodically check for new image files in the ramdrive
        void updateLogic(void);

    private:
        //Runs the streamer program
        QProcess *streamerProc;

        //command-line arguments to be sent to streamer
        QStringList streamerArgs;

        //the image files in the ramdrive
        std::vector <std::string> imageNames;

        //Periodially calls updateLogic
        QTimer logicTimer;

        //the ramdrive directory for this QStreamer
        QDir imageDir;

        //Save the name of the most recent image in the ramdrive. We'll wait a few
        //milliseconds before reading it to give streamer time to finish writing it.
        QString targetImage;

        //location of the video device to use, eg /dev/video0
        QString deviceName;

        //location of the ramdrive to use, must be an absolute location
        QString ramdriveDir;

        //a subdirectory inside of ramdriveDir where we'll read camera images
        QString streamerDir;

        //this is either "ppm" or "jpeg"
        QString fileType;

        //a resolution with the format: "320x240"
        QString resolution;

        //the name of the streamer program to run
        QString programName;

        //a desription of any errors that occur.
        QString errMsg;

        //the number of frames per second to capture
        int framesPerSecond;

        //a unique identifier for this QStreamer instance
        int myID;

        //creates a unique subdirectory in the ramdrive for each QStreamer instance
        static int dirID;

        //True if we started the streamer program, and need to terminate it
        bool streamerStarted;

        //True if targetImage contains an image we should try to load, false if we already loaded it
        bool haveTargetImage;

        //############### FUNCTIONS ###############

        //make a subdirectory in the ramdrive to store images in, if one doesn't already exist
        bool makeSubdirectory(void);

        //Loads the image from file and emits the appropriate signals. Returns false on failure.
        bool loadImage(const QString &imageFile);

};//QStreamer

#endif
